# CP2K


## Summary Version

1.0

## Purpose of Benchmark

CP2K is a freely available quantum chemistry and solid-state physics software package that can perform atomistic simulations of solid state, liquid, molecular, periodic, material, crystal, and biological systems. 

## Characteristics of Benchmark

CP2K can be used to perform DFT calculations using the QuickStep algorithm. This applies mixed  Gaussian and plane waves approaches (such as GPW and GAPW). Supported theory levels include DFTB, LDA, GGA, MP2, RPA, semi-empirical methods (AM1, PM3, PM6, RM1, MNDO, …), and classical force fields (AMBER, CHARMM, …). CP2K can do simulations of molecular dynamics, metadynamics, Monte Carlo, Ehrenfest dynamics, vibrational analysis, core level spectroscopy, energy minimisation, and transition state optimisation using NEB or dimer method.

CP2K is written in Fortran 2008 and can be run in parallel using a combination of multi-threading, MPI, and CUDA. All of CP2K is MPI parallelised, with some additional loops also being OpenMP parallelised. It is therefore most important to take advantage of MPI parallelisation, however running one MPI rank per CPU core often leads to memory shortage. At this point OpenMP threads can be used to utilise all CPU cores without suffering an overly large memory footprint. The optimal ratio between MPI ranks and OpenMP threads depends on the type of simulation and the system in question. CP2K supports CUDA, allowing it to offload some linear algebra operations including sparse matrix multiplications to the GPU through its DBCSR acceleration layer. FFTs can optionally also be offloaded to the GPU. Benefits of GPU offloading may yield improved performance depending on the type of simulation and the system in question.

## Mechanics of Building Benchmark

GNU make and Python 2.x are required for the build process, as are a Fortran 2003 compiler and matching C compiler, e.g. gcc/gfortran (gcc >=4.6 works, later version is recommended).

CP2K can benefit from a number of external libraries for improved performance. It is advised to use vendor-optimized versions of these libraries. If these are not available on your machine, there exist freely available implementations of these libraries including but not limited to those listed below.


<!--- (CP2K is built using a Fortran 2003 compiler and matching C compiler such as gfortran/gcc (version 4.6 and above) and compiled with GNU make. CP2K makes use of a variety of different libraries. Some are essential for running in parallel and others may be used to increase the performance. The steps to build CP2K are as follows:) -->

### Download the source code

Download a CP2K release from https://sourceforge.net/projects/cp2k/files/ or follow instructions at https://www.cp2k.org/download to check out the relevant branch of the CP2K GitHub repository.

### Install or locate required libraries 
**LAPACK & BLAS**

    Can be provided from:

    netlib	                                                                	: http://netlib.org/lapack & http://netlib.org/blas
    MKL	                                                                    	: part of the Intel MKL installation
    LibSci                                                              		: installed on Cray platforms
    ATLAS	                                                                	: http://math-atlas.sf.net
    OpenBLAS                                                                	: http://www.openblas.net
    clBLAS	                                                                        : http://gpuopen.com/compute-product/clblas/

**SCALAPACK and BLACS**

    Can be provided from:

    netlib	                                                                	: http://netlib.org/scalapack/
    MKL	                                                                    	: part of the Intel MKL installation
    LibSci	                                                                   	: installed on Cray platforms

**LIBINT**

Available from - https://www.cp2k.org/static/downloads/libint-1.1.4.tar.gz

The following commands will uncompress and install the LIBINT library required for the UEABS benchmarks:

    tar xzf libint-1.1.4
    cd libint-1.1.4
    ./configure CC=cc CXX=CC --prefix=install_path 	                        	 : must not be this directory
    make
    make install

Note: The environment variables ``CC`` and ``CXX`` are optional and can be used to specify the C and C++ compilers to use for the build (the example above is configured to use the compiler wrappers ``cc`` and ``CC`` used on Cray systems).

### Install optional libraries 
 

    FFTW3	                                                                	: http://www.fftw.org or provided as an interface by MKL
    Libxc	                                                                	: http://www.tddft.org/programs/octopus/wiki/index.php/Libxc
    ELPA	                                                                   	: https://www.cp2k.org/static/downloads/elpa-2016.05.003.tar.gz
    libgrid	                                                                	: within CP2K distribution - cp2k/tools/autotune_grid
    libxsmm	                                                                        : https://www.cp2k.org/static/downloads/libxsmm-1.4.4.tar.gz

### Compile CP2K

Before compiling the choice of compilers, the library locations and compilation and linker flags need to be specified. This is done in an arch (architecture) file. Example arch files for a number of common architecture examples can be found inside the ``cp2k/arch`` directory. The names of these files match the pattern architecture.version (e.g., Linux-x86-64-gfortran.sopt). The case "version=psmp" corresponds to the hybrid MPI + OpenMP version that you should build to run the UEABS benchmarks. Machine specific examples can be found in the relevent subdirectory.

In most cases you need to create a custom arch file, either from scratch or by modifying an existing one that roughly fits the cpu type, compiler, and installation paths of libraries on your system. You can also consult https://dashboard.cp2k.org, which provides sample arch files as part of the testing reports for some platforms (click on the status field for a platform, and search for 'ARCH-file' in the resulting output).

As a guide for GNU compilers the following should be included in the ``arch`` file:

**Specification of which compiler and linker commands to use:**

    CC = gcc
    FC = mpif90
    LD = mpif90
    AR = ar -r

CP2K is primarily a Fortran code, so only the Fortran compiler needs to be MPI-enabled.

**Specification of the ``DFLAGS`` variable, which should include:**

	-D__parallel 	                                                        	: to build parallel CP2K executable)
	-D__SCALAPACK 	                                                                : to link SCALAPACK
	-D__LIBINT 		                                                       	: to link to LIBINT
	-D__MKL 		                                                       	: if relying on MKL for ScaLAPACK and/or an FFTW interface
	-D__HAS_NO_SHARED_GLIBC                                                         : for convenience on HPC systems, see INSTALL.md file

Additional DFLAGS which are needed to link to performance libraries, such as -D__FFTW3 to link to FFTW3, are listed in the INSTALL file. 

**Specification of compiler flags ``FCFLAGS`` (for gfortran):**

    FCFLAGS = $(DFLAGS) -ffree-form -fopenmp                                        : Required
    FCFLAGS = $(DFLAGS) -ffree-form -fopenmp -O3 -ffast-math -funroll-loops         : Recommended

If you want to link any libraries containing header files you should pass the path to the directory containing these to FCFLAGS in the format -I/path_to_include_dir.


**Specification of libraries to link to:**

    -L{path_to_libint}/lib -lderiv -lint                                             : Required for LIBINT

If you use MKL to provide ScaLAPACK and/or an FFTW interface the LIBS variable should be used to pass the relevant flags provided by the MKL Link Line Advisor (https://software.intel.com/en-us/articles/intel-mkl-link-line-advisor), which you should use carefully in order to generate the right options for your system.



#### Building the executable

To build the hybrid MPI+OpenMP executable ``cp2k.psmp`` using *your_arch_file.psmp* run make in the ``cp2k/makefiles`` directory for v4-6 (or in the top-level cp2k directory for v7+).
 
    make -j N ARCH=your_arch_file VERSION=psmp	                               	: on N threads
    make ARCH=your_arch_file VERSION=psmp	                                    	: serially

The executable ``cp2k.psmp`` will then be located in:

    cp2k/exe/your_arch_file



### Compiling CP2K for CUDA enabled GPUs

Arch files for compiling CP2K for CUDA enabled GPUs can be found here:

In general the main steps are:

1. Load the cuda module.
2. Ensure that CUDA_PATH variable is set.
3. Add the following to the arch file:


**Addtional required compiler and linker commands**

    NVCC = nvcc

**Additional ``DFLAGS``**

    -D__ACC -D__DBCSR_ACC -D__PW_CUDA
 
**Set ``NVFLAGS``**

    NVFLAGS = $(DFLAGS) -O3 -arch sm_60
 
**Additional required libraries**
 
    -lcudart -lcublas -lcufft -lrt 



## Mechanics of Running Benchmark

The general way to run the benchmarks with the hybrid parallel executable is:

    export OMP_NUM_THREADS=X   
    parallel_launcher launcher_options path_to_/cp2k.psmp -i inputfile.inp -o logfile  

Where:

* The environment variable for the number of threads must be set before calling the executable.
* The parallel_launcher is mpirun, mpiexec, or some variant such as aprun on Cray systems or srun when using Slurm.
* launcher_options specifies parallel placement in terms of total numbers of nodes, MPI ranks/tasks, tasks per node, and OpenMP threads per task (which should be equal to the value given to OMP_NUM_THREADS). This is not necessary if parallel runtime options are picked up by the launcher from the job environment.
* You can try any combination of tasks per node and OpenMP threads per task to investigate absolute performance and scaling on the machine of interest.
* The inputfile usually has the extension .inp, and may specify within it further requried files (such as basis sets, potentials, etc.)

You can try any combination of tasks per node and OpenMP threads per task to investigate absolute performance and scaling on the machine of interest. For tier-1 systems the best performance is usually obtained with pure MPI, while for tier-0 systems the best performance is typically obtained using 1 MPI task per node with the number of threads being equal to the number of cores per node.

**UEABS benchmarks:**

Test Case | System     | Number of Atoms | Run type      | Description                                          | Location                        |
----------|------------|-----------------|---------------|------------------------------------------------------|---------------------------------|
a         | H2O-512    |   1236          | MD            | Uses the Born-Oppenheimer approach via Quickstep DFT | ``/tests/QS/benchmark/``        |
b         | LiHFX      |      216        | Single-energy | Must create wavefuntion first - see benchmark README | ``/tests/QS/benchmark_HFX/LiH`` |
c         | H2O-DFT-LS | 6144            | Single-energy | Uses linear scaling DFT                              | ``/tests/QS/benchmark_DM_LS``    |
 
More information in the form of a README and an example job script is included in each benchmark tar file.

## Verification of Results

The run walltime is reported near the end of logfile:

    grep "CP2K    " logfile | awk -F ' ' '{print $7}'

